import h5py as _h5
import pathlib
import numpy as np

from ..traces.abstract_reader import AbstractReader
from ..traces.trace_header_set import build_trace_header_set
from . import frames


def read_ths_from_ets_file(filename):
    """Build and returns a :class:`TraceHeaderSet` instance from an ETS file.

    Args:
        filename (str or Path): filename of the ETS file

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    return build_trace_header_set(
        reader=ETSFormatReader(filename=filename),
        name='ETS Format THS'
    )


_TRACES_KEY = 'traces'
_METADATAS_KEY = 'metadata'


def _cast_to_str(val):
    # h5py doesn't manage nicely strings values. We must cast these explicitly before returning to the user.
    if isinstance(val, np.ndarray):
        if 'S' in str(val.dtype):
            return np.array([r.decode() for r in val])
    if isinstance(val, bytes):
        return val.decode()
    return val


class ETSFormatReader(AbstractReader):

    def __init__(self, filename):
        self._filename = str(filename)
        if not pathlib.Path(filename).exists():
            raise AttributeError(f'File {filename} does not exist.')
        if not _h5.h5f.is_hdf5(bytes(self._filename, 'ascii')):
            raise TypeError(f'File {self._filename} is not a valid ETS file.')

        self._file = None
        self._sub_traceset_indices = None
        self._size, self._trace_size = self._h5_file[_TRACES_KEY].shape

    @property
    def _metadatas(self):
        return self._h5_file[_METADATAS_KEY]

    def fetch_metadatas(self, key, trace_id):
        if trace_id is not None:
            trace_index = self._convert_trace_index_to_file_index(trace_id)
            res = self._metadatas[key][trace_index]
        else:
            if isinstance(self._traceset_indices, slice):
                res = self._metadatas[key][self._traceset_indices]
            else:
                res = np.array([self._metadatas[key][i] for i in self._traceset_indices])
        return _cast_to_str(res)

    @property
    def _traceset_indices(self):
        if self._sub_traceset_indices is not None:
            if frames.is_array_equivalent_to_a_slice(self._sub_traceset_indices):
                return frames.build_equivalent_slice(array=self._sub_traceset_indices)
            return self._sub_traceset_indices
        return slice(None, None, 1)

    def _convert_trace_index_to_file_index(self, trace_id):
        return self._convert_traces_indices_to_file_indices_array([trace_id])[0]

    def _convert_traces_indices_to_file_indices_key(self, traces):
        indices = self._convert_traces_indices_to_file_indices_array(traces=traces)
        if frames.is_array_equivalent_to_a_slice(indices):
            return frames.build_equivalent_slice(array=indices)
        return indices

    def _convert_traces_indices_to_file_indices_array(self, traces):
        if self._sub_traceset_indices is not None:
            sub_max = len(self._sub_traceset_indices)
            traces_index = np.array([t for t in traces if t < sub_max])
            return self._sub_traceset_indices[traces_index]
        return np.array(traces)

    def fetch_samples(self, traces, frame):
        traces = self._convert_traces_indices_to_file_indices_key(traces)

        if isinstance(frame, int):
            frame = np.array([frame])
        elif isinstance(frame, list):
            frame = np.array(frame)
        if isinstance(frame, np.ndarray) and frames.is_array_equivalent_to_a_slice(frame):
            frame = frames.build_equivalent_slice(frame)

        if isinstance(frame, (Ellipsis.__class__, slice)):
            return self._slice_traces(keys_tuple=(traces, frame))
        else:
            start = np.min(frame)
            stop = np.max(frame) + 1
            frame = frame - start
            return self._slice_traces(keys_tuple=(traces, slice(start, stop)), final_frame=frame)
            # TODO: how to write a test case for the memory overhead optimization strategy below ?
            # ratio = (stop - start) / frame.size
            # if ratio > 100:
            #   sorted_deduped_frame, inversion_frame = np.unique(frame, return_inverse=True)
            #   return self._slice_traces(keys_tuple=(traces, sorted_deduped_frame), final_frame=inversion_frame)

    def _slice_traces(self, keys_tuple, final_frame=None):
        traces, frame_slice = keys_tuple
        if final_frame is not None:
            if isinstance(traces, slice):
                return self._h5_file[_TRACES_KEY][keys_tuple][:, final_frame]
            if len(traces) > 0:
                return np.vstack([self._h5_file[_TRACES_KEY][t, frame_slice][final_frame] for t in traces])
        else:
            if isinstance(traces, slice):
                return self._h5_file[_TRACES_KEY][keys_tuple]
            if len(traces) > 0:
                return np.vstack([self._h5_file[_TRACES_KEY][t, frame_slice] for t in traces])
        return np.array([])

    def __getitem__(self, key):
        super().__getitem__(key)

        if isinstance(key, int):
            key = [key]
        elif isinstance(key, slice):
            key = range(
                key.start if key.start is not None else 0,
                key.stop if key.stop is not None else len(self),
                key.step if key.step is not None else 1
            )
        sub_traceset_indices = self._convert_traces_indices_to_file_indices_array(traces=key)
        traces_number = len(sub_traceset_indices)
        new_reader = ETSFormatReader(filename=self._filename)
        new_reader._file = self._file
        new_reader._sub_traceset_indices = sub_traceset_indices
        new_reader._size = traces_number
        return new_reader

    @property
    def metadatas_keys(self):
        # A keys only dict value is instantiated here because h5py doesn't manage properly all keys' types.
        return {k: None for k in self._h5_file[_METADATAS_KEY].keys()}.keys()

    def get_trace_size(self, trace_id):
        return self._trace_size

    @property
    def _h5_file(self):
        if self._file is None:
            self._file = _h5.File(name=self._filename, mode='r', libver='latest', swmr=True)
        return self._file

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(filename={self._filename})'
        )

    def __str__(self):
        return f'ETS format reader of file {self._filename} with {len(self)} traces.'



# -*- coding: utf-8 -*-
from ..formats import ets_format as _ets_format
import warnings as _warnings


def _ets_reader(filename):
    _warnings.warn(
        'ETSReader is deprecated. Use read_ths_from_ets functions from estraces package.',
        DeprecationWarning
    )

    return _ets_format.read_ths_from_ets_file(filename=filename)


ETSReader = _ets_reader



# -*- coding: utf-8 -*-
from . import metadatas
from . import samples


class Trace:
    """Provides a consistent API to manipulate a trace samples data and metadatas.

    Attributes:
        id (int): identifier of the trace
        name (str): name of the trace
        samples (:class:`Samples`): 1 dimension samples data
        metadatas (:class:`Metadatas`): trace metadatas

    Note:
        All metadatas are available through the `metadatas` attributes and through a corresponding named property.

    Examples:
        Samples are available as 1 dimensionnal array-like :class:`Samples` with shape (size of trace,).
        It supports a subset of numpy.ndarray-like slicing - including advanced list-based slicing::

            trace.samples[2000:3000]
            trace.samples[[1, 100, 1000]] # get samples at 1, 100 and 1000 indexes

        Metadatas are available through a dict-like :class:`Metadatas` instance::

            metadatas = trace.metadatas
            metadatas['plaintext']

        Each metadata can be reached with its own property::

            trace.plaintext # is equivalent to trace.metadatas['plaintext']

    """

    def __init__(self, trace_id, reader):
        if trace_id is None:
            raise AttributeError("trace_id can't be None.")
        self.id = trace_id
        self._reader = reader
        self._samples = None
        self._metadatas = None
        self.name = f'Trace n°{self.id}'

    def __len__(self):
        return len(self.samples)

    @property
    def samples(self):
        if self._samples is None:
            self._samples = samples.Samples(reader=self._reader, trace_id=self.id)
        return self._samples

    @property
    def metadatas(self):
        if not self._metadatas:
            self._metadatas = metadatas.Metadatas(reader=self._reader, trace_id=self.id)
        return self._metadatas

    def __repr__(self):
        return f'{self.__class__.__name__}(trace_id={self.id}, reader={self._reader})'

    def __str__(self):
        r = 'Trace\n'
        r += f'{"Id":.<17}: {self.id}\n'
        r += f'{"Samples size":.<17}: {len(self)}\n'
        for k in self.metadatas.keys():
            r += f'{k:.<17}: {self.metadatas.get(k)}\n'
        return r



from .bin_format import read_ths_from_bin_filenames_list, read_ths_from_bin_filenames_pattern
from .ets_format import read_ths_from_ets_file
from .trs_format import read_ths_from_trs_file
from .ram_format import read_ths_from_ram

__all__ = [
    "read_ths_from_bin_filenames_list",
    "read_ths_from_bin_filenames_pattern",
    "read_ths_from_ets_file",
    "read_ths_from_trs_file",
    "read_ths_from_ram"
]



import numpy as np


def is_array_equivalent_to_a_slice(array):
    return array is not None and _array_has_enough_items(array) and _items_are_in_ascending_order(array) and _items_are_equally_spaced(array)


def _items_are_in_ascending_order(array):
    return np.any(np.diff(array) > 0)


def _array_has_enough_items(array):
    return len(array) >= 3


def _items_are_equally_spaced(array):
    return not np.any(np.diff(array, 2))


def build_equivalent_slice(array):
    if len(array) == 0:
        return slice(0, 0, 1)
    start = np.min(array)
    stop = np.max(array) + 1
    step = array[1] - array[0] if len(array) > 1 else 1
    return slice(start, stop, step)



# -*- coding: utf-8 -*-
"""Provides metaclass for Trace and TraceHeaderSet classes.

:copyright: (c) 2015-2019 ESHARD
"""


class ClassWithMetadatas(type):
    """Metaclass used to create classes providing attribute-like access to metadatas associated to a primary data sourced wrapped by the class.

    It is used to create specific classes inheriting from :class:`Trace` class or from :class:`TraceHeaderSet` class, with attribute accessor to metadatas
    suited to the specific Format implementation provided.
    """

    def __new__(cls, name, bases, namespace, metadatas_keys):
        res = type.__new__(cls, name, bases, namespace)
        res._metadatas_keys = metadatas_keys
        res._set_metadatas_properties()
        res._is_valid_trace_class = True
        res.__doc__ = bases[0].__doc__
        return res

    def _set_metadatas_properties(cls):
        for k in cls._metadatas_keys:
            setattr(cls, f"{k}", cls._create_metadata_property(metadata_key=k))

    def _create_metadata_property(cls, metadata_key):

        def _(self):
            return self.metadatas[metadata_key]

        return property(fget=_, doc="""Returns f'{metadata_key}' array of values.""")



import abc
import typing
import numpy as _np

"""Abtract base class for concrete trace formats.


"""

FrameType = typing.Union[....__class__, int, slice, list, _np.ndarray]
MetaType = typing.TypeVar("M")
MetadataTypes = typing.Union[MetaType, typing.Container[MetaType]]
TraceId = typing.TypeVar("I")


class AbstractReader(abc.ABC):
    """Provides the basic interface any format reader for trace set must provides.

    The :class:`TraceHeaderSet` and :class:`Trace` abstraction rely on the use of concrete implementation of `AbstractReader`.
    Format implementation must provides fetching methods for samples and metadatas, and must implement `__getitem__`.
    All performances strategies (lazy loading, caching, memory or CPU usage) are thus delegated to the implementer of the concrete format.

    Methods:
        fetch_samples: given traces and samples frame, returns samples data
        fetch_metadatas: given a trace id and a metadata key, returns metadata values
        __getitem__: handles slicing of reader
        metadatas_keys: property - Returns a view or container of the metadatas available keys

    """

    _size = None

    def __len__(self):
        return self._size

    @abc.abstractmethod
    def fetch_samples(self, traces: list, frame=None) -> _np.ndarray:
        """Fetch samples for the given traces id and given samples data frame.

        Args:
            traces: Lists of traces id to fetch.
            frame: Samples data to fetch. Must support `Ellipsis`, `slice`, `list`, `ndarray` or `int` types.

        Returns:
            (:class:`numpy.ndarray`) array of shape (number of traces, size of samples)

        """
        pass

    @abc.abstractmethod
    def fetch_metadatas(self, key: typing.Hashable, trace_id: int = None) -> MetadataTypes:
        """Fetch metadata value for the given metadata key and trace id.

        Args:
            key (typing.Hashable): Key of the metadata to fetch. Must be hashable.
            trace_id (int): Trace id for which to fetch the metadata.

        Returns:
            A container of all the values of the trace set for the given metadata if trace_id is None.
            Else, the value of the metadata for the given trace id.

        """
        pass

    @abc.abstractmethod
    def __getitem__(self, key):
        """Returns a new format instance limited to traces[key] subset.

        Args:
            key: slice or list of traces indexes to slice on.

        """
        if not isinstance(key, (slice, list, _np.ndarray)):
            raise TypeError('Only slice, 1 dimension numpy array and lists are valid indices for types implementing AbstractReader')

    @property
    @abc.abstractmethod
    def metadatas_keys(self):
        """Provides a list or views of the metadatas keys available."""
        pass

    @abc.abstractmethod
    def get_trace_size(self, trace_id):
        """Provides the size of trace trace_id."""
        pass



import warnings


class TraceDeprecated:

    @property
    def nb_point(self):
        warnings.warn(
            "Trace.nb_point method is deprecated. Use len(trace) instead.",
            DeprecationWarning
        )
        return len(self)

    def points(self, frame=None):
        warnings.warn(
            "Trace.points method is deprecated. Use .samples attribute instead.",
            DeprecationWarning
        )
        if frame is None:
            return self.samples[:]
        if isinstance(frame, range):
            frame = slice(frame.start, frame.stop, frame.step)
        if isinstance(frame, tuple):
            frame = list(frame)
        return self.samples[frame]

    @property
    def filename(self):
        try:
            return self._reader._filenames[self.id]
        except Exception:
            return AttributeError('filename attribute does not exist.')

    def __get_writable_attributes__(self):
        res = list(self.metadatas.keys())
        for k, v in self.__dict__.items():
            if k not in ('points', 'filename', 'nb_point', 'samples', 'metadatas', 'id', 'name') and k[0] != '_' and k[0:1] != '__':
                res.append(k)
        return res



# -*- coding: utf-8 -*-
import os as _os
import glob as _glob
import re as _re
import numpy as _np
import enum as _enum
from . import frames
from . import bin_extractor
from ..traces.trace_header_set import build_trace_header_set
from ..traces.abstract_reader import AbstractReader

__all__ = [
    'get_sorted_filenames',
    'read_ths_from_bin_filenames_list',
    'read_ths_from_bin_filenames_pattern',
    'PaddingMode',
    'BinFormat'
]


def get_sorted_filenames(pattern):
    """Retrieve a filename list conforming to a given glob string pattern, sorted by digit chunk found in filename."""
    files = _glob.glob(pattern)
    if not files:
        raise FileNotFoundError("No file found with pattern: '" + pattern + "'")
    files = sorted(
        files,
        key=lambda ll: [
            int(chunk) if chunk.isdigit() else 0 for chunk in _re.split("([0-9]+)", ll)
        ],
    )
    return files


class PaddingMode(_enum.Enum):
    """Defines the padding mode used when traces of binary format have different length.

    Possible modes are:
        - NONE: no padding is applied, an exception is thrown when instantiating the trace set.
        - PAD: samples will be padded with zero to the length of the longer trace of the set.
        - TRUNCATE: samples will be truncated to the length of the smaller trace of the set.
    """

    NONE = 0
    PAD = 1
    TRUNCATE = 2


def read_ths_from_bin_filenames_pattern(filename_pattern, metadatas_parsers, dtype, offset=0, padding_mode=PaddingMode.NONE):
    """Build and returns a :class:`TraceHeaderSet` instance from binaries files following a filename pattern.

    Each file must contains samples data for one trace.

    Args:
        filename_pattern (str): binary files filename pattern, for example ``"samples/*.bin"``
        metadatas_parsers (dict): dict with a key for each metadata of the trace set, with value a :class:`bin_extractor.HeaderExtractor`,
            :class:`bin_extractor.PatternExtractor`, :class:`bin_extractor.FilePatternExtractor` or :class:`bin_extractor.DirectValue` instance.
        dtype (dtype): the Numpy samples data dtype
        offset (int, default:0): use as an offset when reading samples in files.
        padding_mode (:class:`PaddingMode`, default: `PaddingMode.NONE`): choose how to handle different traces size in your trace list.
            Possible modes are NONE, PAD and TRUNCATE (see :class:`bin_format.PaddingMode`).

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    files_list = get_sorted_filenames(pattern=filename_pattern)
    return read_ths_from_bin_filenames_list(
        filenames_list=files_list,
        metadatas_parsers=metadatas_parsers,
        dtype=dtype,
        offset=offset,
        padding_mode=padding_mode
    )


def read_ths_from_bin_filenames_list(filenames_list, metadatas_parsers, dtype, offset=0, padding_mode=PaddingMode.NONE):
    """Build and returns a :class:`TraceHeaderSet` instance from binaries files listed.

    Each file must contains samples data for one trace.

    Args:
        filenames_list (list): binary files filenames list
        metadatas_parsers (dict): dict with a key for each metadata of the trace set, with value a :class:`bin_extractor.HeaderExtractor`,
            :class:`bin_extractor.PatternExtractor`, :class:`bin_extractor.FilePatternExtractor` or :class:`bin_extractor.DirectValue` instance.
        dtype (dtype): the Numpy samples data dtype
        offset (int, default: 0): use as an offset when reading samples in files.
        padding_mode (:class:`PaddingMode`, default: `PaddingMode.NONE`): choose how to handle different traces size in your trace list.
            Possible values are NONE, PAD and TRUNCATE (see :class:`bin_format.PaddingMode`).

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    return build_trace_header_set(
        reader=BinFormat(
            filenames=filenames_list,
            offset=offset,
            dtype=dtype,
            metadatas_parsers=metadatas_parsers,
            padding_mode=padding_mode
        ),
        name="BinFormat trace header set",
    )


class _FExtractor:

    def __init__(self, filename, index_conversion):
        self._content = None
        self._filename = filename
        self._index_conversion = index_conversion

    def __call__(self, index):
        if self._content is None:
            with open(self._filename, 'r') as f:
                self._content = f.readlines()
        if len(self._content) == 1:
            return self._content[0]
        else:
            return self._content[self._index_conversion(index)]


class BinFormat(AbstractReader):

    def __init__(self, filenames, dtype, metadatas_parsers, offset=0, padding_mode=PaddingMode.NONE, metadatas_indices_conversion_function=lambda i: i):
        if not isinstance(filenames, list):
            raise TypeError("filenames must be a list of filenames strings.")

        self._metadatas_indices_conversion_function = metadatas_indices_conversion_function
        self._filenames = _np.array(filenames)
        self._raw_metas = metadatas_parsers
        self._dtype = _np.dtype(dtype)
        self._offset = offset
        self._set_padding_mode(padding_mode)
        self._initialize_metadatas(metadatas_parsers)

    def _set_padding_mode(self, padding_mode):
        if not isinstance(padding_mode, PaddingMode):
            raise AttributeError(f'padding_mode must be a PaddingMode enum instance, not {type(padding_mode)}.')
        self._padding_mode = padding_mode

        lens = _np.array([self._trace_file_size(i) for i in range(len(self))])
        if self._padding_mode == PaddingMode.NONE:
            try:
                self._trace_size = lens[0]
            except IndexError:
                self._trace_size = 0
            if _np.any(_np.diff(lens)):
                raise ValueError(
                    f'Not all traces are of the same length, and you are using no padding mode of BinFormat. \
                        You should either fix you trace set files or use TRUNCATE or PAD padding mode.')
        elif self._padding_mode == PaddingMode.TRUNCATE:
            self._trace_size = lens.min()
        elif self._padding_mode == PaddingMode.PAD:
            self._trace_size = lens.max()

    def _initialize_metadatas(self, raw_metas):
        self._metadatas_parsers = {}

        def _2(index):
            return self._filenames[index]

        for k, extractor in raw_metas.items():
            if isinstance(extractor, bin_extractor.FilePatternExtractor):
                p = bin_extractor.PatternExtractor(
                    pattern=extractor.pattern,
                    replace=extractor.replace,
                    num=extractor.num,
                    unhexlify=extractor.unhexlify
                )

                ex = _FExtractor(filename=extractor.filename, index_conversion=self._metadatas_indices_conversion_function)
                self._metadatas_parsers[k] = (p, ex)
            else:
                self._metadatas_parsers[k] = (extractor, _2)

    def __len__(self):
        return len(self._filenames)

    def __getitem__(self, key):
        super().__getitem__(key)
        if isinstance(key, (list, _np.ndarray)):
            filenames = [self._filenames[i] for i in key]
        elif isinstance(key, slice):
            filenames = self._filenames[key].tolist()

        indices_map = {
            new_index: self._filenames.tolist().index(filenames[new_index]) for new_index in range(len(filenames))
        }

        fmt = BinFormat(
            filenames=filenames,
            offset=self._offset,
            dtype=self._dtype,
            metadatas_parsers=self._raw_metas,
            metadatas_indices_conversion_function=lambda i: self._metadatas_indices_conversion_function(indices_map[i]),
            padding_mode=self._padding_mode
        )
        return fmt

    def _read_samples(self, filename, frame):
        if isinstance(frame, int):
            frame = _np.array([frame])
        elif isinstance(frame, list):
            frame = _np.array(frame)

        if isinstance(frame, _np.ndarray) and frames.is_array_equivalent_to_a_slice(frame):
            frame = frames.build_equivalent_slice(frame)

        readsize = self._trace_size * self._dtype.itemsize
        with open(filename, "rb") as trace_file:
            trace_file.seek(self._offset)
            samples = _np.frombuffer(trace_file.read(readsize), dtype=self._dtype)

        if self._padding_mode == PaddingMode.PAD and samples.shape[0] < self._trace_size:
            t_samples = _np.zeros((self._trace_size), dtype=self._dtype)
            t_samples[:samples.shape[0]] = samples
            return t_samples[frame].squeeze()

        return samples[frame].squeeze()

    def fetch_samples(self, traces, frame):
        if len(traces) == 0 or len(self._filenames) == 0:
            return _np.array([], dtype=self._dtype)
        return _np.vstack(
            [
                self._read_samples(filename=self._filenames[i], frame=frame)
                for i in traces
            ]
        )

    def fetch_metadatas(self, key, trace_id=None):
        if trace_id is not None:
            return _apply_extractor(
                extractor=self._metadatas_parsers[key][0],
                value_function=self._metadatas_parsers[key][1],
                index=trace_id
            )
        return _np.array(
            [
                _apply_extractor(
                    extractor=self._metadatas_parsers[key][0],
                    value_function=self._metadatas_parsers[key][1],
                    index=idx
                )
                for idx in range(len(self._filenames))
            ]
        )

    def get_trace_size(self, trace_id):
        return self._trace_size

    def _trace_file_size(self, trace_id):
        file_size = _os.stat(self._filenames[trace_id]).st_size
        return int((file_size - self._offset) / self._dtype.itemsize)

    @property
    def metadatas_keys(self):
        return self._metadatas_parsers.keys()

    @property
    def _filename(self):
        return self._filenames

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(filenames={self._filenames}, dtype={self._dtype}, offset={self._offset}, metadatas_parsers={self._metadatas_parsers})'
        )

    def __str__(self):
        return f'Bin format reader with {len(self._filenames)} files, dtype {self._dtype}'


# TODO: refactor pattern extractor
def _apply_extractor(extractor, value_function, index):
    try:
        if extractor.unhexlify is True:
            return _np.frombuffer(extractor.get_text(value_function(index)), dtype=_np.uint8)
        return extractor.get_text(value_function(index))
    except AttributeError:
        return extractor



from ..traces import abstract_reader
from ..traces.trace_header_set import build_trace_header_set
import numpy as _np
import trsfile as _trsfile


def read_ths_from_trs_file(filename, metadatas_parsers, dtype=None):
    """Build and returns a :class:`TraceHeaderSet` instance from TraceRiscureSet file.

    Args:
        filename (str): TRS filename
        metadatas_parsers (dict): dict with a key for each metadata of the trace set,
            with value a function or lambda taking `data` as input and returning the metadata value parsed from it.
        dtype (dtype, optionnal): a Numpy dtype that will override the TRS samples data type

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    return build_trace_header_set(
        reader=TRSFormatReader(
            filename=filename,
            metadatas_parsers=metadatas_parsers,
            dtype=dtype
        ),
        name="TrsFromat trace header set"
    )


class TRSFormatReader(abstract_reader.AbstractReader):

    def __init__(self, filename, metadatas_parsers, dtype=None):
        self._filename = filename
        self._sub_traceset_indices = None
        try:
            with _trsfile.open(filename, 'r') as trs_file:
                self._headers = trs_file.get_headers()
                self._size = self._headers[_trsfile.Header.NUMBER_TRACES]
        except (FileNotFoundError, ValueError, NotImplementedError) as e:
            raise AttributeError(f"{filename} is not a valid TRS file or filename. Original exception was:{e}.")
        self._set_metadatas_parsers(metadatas_parsers)
        self.dtype = _np.dtype(dtype) if dtype else self._headers[_trsfile.Header.SAMPLE_CODING].format

    def _set_metadatas_parsers(self, metadatas_parsers):
        self._metadatas_parsers = metadatas_parsers
        try:
            for k in self.metadatas_keys:
                self.fetch_metadatas(key=k, trace_id=0)
        except TypeError as e:
            raise AttributeError(f'Metadatas parsers {metadatas_parsers} are not valid unary functions. Original exception was:{e}.')

    def fetch_samples(self, traces, frame=None):
        traces = self._convert_traces_indices_to_file_indices_array(traces)

        if isinstance(frame, int):
            frame = [frame]
        with _trsfile.open(self._filename, 'r') as trs_file:
            raw_traces = [trs_file[trace] for trace in traces]
            samples = _np.array([trace[frame] for trace in raw_traces], dtype=self.dtype)
        return samples

    def fetch_metadatas(self, key, trace_id):
        if trace_id is not None:
            trace_index = self._convert_traces_indices_to_file_indices_array(trace_id)
            with _trsfile.open(self._filename, 'r') as trs_file:
                raw_trace = trs_file[trace_index]
                res = self._metadatas_parsers[key](raw_trace.data)
        else:
            if self._sub_traceset_indices is not None:
                indices = [i for i in self._sub_traceset_indices]
            else:
                indices = [i for i in range(len(self))]

            with _trsfile.open(self._filename, 'r') as trs_file:
                res = _np.array([self._metadatas_parsers[key](trs_file[i].data) for i in indices])
        return res

    def __getitem__(self, key):
        super().__getitem__(key)
        if isinstance(key, int):
            key = [key]
        elif isinstance(key, slice):
            key = range(
                key.start if key.start is not None else 0,
                key.stop if key.stop is not None else len(self),
                key.step if key.step is not None else 1
            )
        sub_traceset_indices = self._convert_traces_indices_to_file_indices_array(traces=key)
        traces_number = len(sub_traceset_indices)
        new_reader = TRSFormatReader(
            filename=self._filename,
            metadatas_parsers=self._metadatas_parsers
        )
        new_reader._sub_traceset_indices = sub_traceset_indices
        new_reader._size = traces_number
        return new_reader

    @property
    def metadatas_keys(self):
        return self._metadatas_parsers.keys()

    def get_trace_size(self, trace_id):
        return self._headers[_trsfile.Header.NUMBER_SAMPLES]

    def _convert_traces_indices_to_file_indices_array(self, traces):
        if self._sub_traceset_indices is not None:
            sub_max = len(self._sub_traceset_indices)
            traces_index = [t for t in traces if t < sub_max]
            return _np.array(self._sub_traceset_indices[traces_index])
        return _np.array(traces)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(filename={self._filename}, metadatas_parsers={self._metadatas_parsers}, dtype={self.dtype})'
        )

    def __str__(self):
        return f'TRS format reader with {self._filename}, contains {len(self)} traces.'



import os as _os
import warnings as _warn
import h5py as _h5py
import numpy as _np
from .ets_format import read_ths_from_ets_file


class ETSWriter:
    """Class used to write a eshard custom format trace set."""

    def __init__(self, filename, overwrite=False):
        self._overwrite = overwrite
        self._filename = filename
        if _os.path.isfile(self._filename):
            if self._overwrite:
                _warn.warn("File '%s' already exists. It will be reseted on the first write operation" % self._filename)
            else:
                _warn.warn("File '%s' already exists. If no indexes are specified, new data will be append on existing datasets" % self._filename)
        self._isinit = False

    def __close_all__(self):
        try:
            f = _h5py.File(self._filename)
            to_close = _h5py.h5f.get_obj_ids(f.id)
        except Exception:
            to_close = []
        for fid in to_close:
            tmp = _h5py.File(fid)
            tmp.flush()
            tmp.close()
        try:
            f.close()
        except Exception:
            pass

    def __init_file__(self):
        self.__close_all__()
        if self._overwrite:
            _os.remove(self._filename)
            self._h5_file = _h5py.File(self._filename, mode='w', libver='latest')
        else:
            self._h5_file = _h5py.File(self._filename, mode='a', libver='latest')
        try:
            self._h5_file.swmr_mode = True
        except Exception:
            pass
        if 'metadata' not in list(self._h5_file.keys()):
            self._h5_file.create_group('metadata')
        self._initialized_datasets = list(self._h5_file['metadata'].keys())
        self._is_init_traces = 'traces' in list(self._h5_file.keys())
        self._isinit = True

    def __del__(self):
        self.close()

    @staticmethod
    def __np_to_str__(data):
        if data.shape == ():
            data = str(data)
        else:
            data = str(data)
            data = data[1:-1]
        return data

    def write_trace_object_and_points(self, trace_object, points, index=None):
        try:
            writable_attributes = trace_object.__get_writable_attributes__()
        except Exception:
            writable_attributes = trace_object.__dict__.keys()
        for tag in writable_attributes:
            tmp = getattr(trace_object, tag)
            self.write_meta(tag=tag, metadata=tmp, index=index)
        self.write_points(points=points, index=index)

    def write_points(self, points, index=None):
        if not self._isinit:
            self.__init_file__()
        if not isinstance(points, _np.ndarray) or points.ndim != 1:
            raise Exception("'points' should be a 1D ndarray")
        if not self._is_init_traces:
            self.__init_traces__(points)
        self.__add_to_dataset__(self._h5_file['traces'], points, index)

    def write_meta(self, tag, metadata, index=None):
        if not self._isinit:
            self.__init_file__()
        if not isinstance(metadata, _np.ndarray):
            metadata = _np.array(metadata)
        metadata = metadata.squeeze()

        if tag not in self._initialized_datasets:
            self.__init_meta__(tag, metadata)
        self.__add_to_dataset__(self._h5_file['metadata'][tag], metadata, index)

    def __init_traces__(self, points):
        points = points.squeeze()
        try:
            trace_size = len(points)
        except TypeError:
            trace_size = 1
        dtype = points.dtype
        self._h5_file.create_dataset('traces', (0, trace_size), maxshape=(None, trace_size), dtype=dtype)
        self._is_init_traces = True

    def __init_meta__(self, tag, first_meta):
        if first_meta.dtype.kind not in ['b', 'i', 'u', 'f', 'c']:
            h5type = _h5py.special_dtype(vlen=str)
            self._h5_file['metadata'].create_dataset(tag, (0,), maxshape=(None, ), dtype=h5type)
            self._initialized_datasets.append(tag)
        else:
            try:
                meta_size = len(first_meta)
            except TypeError:
                meta_size = 1
            dtype = first_meta.dtype
            self._h5_file['metadata'].create_dataset(tag, (0, meta_size), maxshape=(None, meta_size), dtype=dtype)
            self._initialized_datasets.append(tag)

    def __add_to_dataset__(self, data_set, data, index):
        try:
            if data_set.dtype.kind != 'O':
                expected_size = data_set.shape[-1]
                try:
                    data_size = len(data)
                except TypeError:
                    data_size = 1
                if data_size == expected_size:
                    tmp_data = data
                else:
                    tmp_data = _np.empty(expected_size, dtype=data.dtype)
                    data = data[:expected_size]
                    tmp_data[:len(data)] = data
                    tmp_data[len(data):] = 0
            else:
                tmp_data = self.__np_to_str__(data)
            nb_data = data_set.shape[0]

            if index is None:
                index = nb_data
            if index < nb_data and not self._overwrite:
                raise Exception("An element already exists in '%s' dataset at index %d and overwriting is disabled" % (str(data_set.name), index))

            if index >= data_set.shape[0]:
                data_set.resize(index + 1, axis=0)
            data_set[index] = tmp_data
        except Exception as e:
            raise e
        finally:
            self._h5_file.flush()

    def get_reader(self):
        self.close()
        return read_ths_from_ets_file(self._filename)

    def close(self):
        try:
            self._h5_file.close()
        except Exception:
            pass



# -*- coding: utf-8 -*-
"""eshard's Trace set Python library.

esTraces provides the basic tools to handle side-channel traces in a unified way, through:
   * classes dedicated to trace and trace set manipulation: :class:`estraces.TraceHeaderSet`, :class:`estraces.Trace`
   * specific file formats reader: binaries formats and Eshard format ETS
   * abstract base class to easily implement a reader for any specific file format

Trace and TraceHeaderSet
========================

:class:`estraces.Trace` and :class:`estraces.TraceHeaderSet` are the classes which abstract your samples and the corresponding metadatas.
Through these objects, you can:
* manipulate your trace set with TraceHeaderSet - filter it, slice it, ...
* manipulate your samples data through the `samples` attribute of both TraceHeaderSet and Trace, which provides a Numpy array-like API.
* manipulate your metadata through the `metadatas` attribute of both TraceHeaderSet and Trace, which provides a dict-like interface to your metadatas.

All the underlying mechanics to read datas on your trace set files are managed by a format reader.

`Trace` class
-------------
.. autoclass:: estraces.Trace

`TraceHeaderSet` class
----------------------
.. autoclass:: estraces.TraceHeaderSet
    :members:
    :member-order: bysource

`Samples` class
---------------
.. autoclass:: estraces.Samples
    :members:
    :member-order: bysource

`Metadatas` class
-----------------
.. autoclass:: estraces.Metadatas
    :members:
    :member-order: bysource


Get a trace header set from files - Format readers
==================================================

Basically, formats reader are classes which implements methods to read samples and metadatas from a trace set files. For the sake of simplicity,
a format should always comes with a simple factory function to get a ths from basic parameters. esTraces provides:

   * a bin format reader, which comes with two factory functions
   * the eshard ETS format reader which comes with its factory functions :func:`estraces.read_ths_from_ets_file`

Bin reader format: `read_ths_from_bin_...` functions
----------------------------------------------------

.. autofunction:: estraces.read_ths_from_bin_filenames_pattern


.. autofunction:: estraces.read_ths_from_bin_filenames_list


.. automodule:: estraces.formats.bin_extractor
    :members:


ETS reader format:  `read_ths_from_ets_file`
--------------------------------------------

.. autofunction:: estraces.read_ths_from_ets_file


Implementing a new format reader
================================

To handle your own trace set format, you must create a format reader class inheriting from :class:`estraces.AbstractReader`.
A reader implementation must basically implement two methods:
* a `fetch_samples` method which takes a traces id list and samples frame as inputs, and returns a 2 dimension numpy array
* a `fetch_metadatas` method which takes a trace id and a metadata name, and returns the metadata values

When developing your own format reader, you should also provide a factory function to get a TraceHeaderSet instance from the files.

The base format reader `AbstractReader`
---------------------------------------

.. autoclass:: estraces.AbstractReader
    :members:
    :member-order: bysource

How to create a `TraceHeaderSet` instance
-----------------------------------------

.. autofunction:: estraces.build_trace_header_set


:copyright: (c) 2019 ESHARD

=========

"""

from .traces import Trace, TraceHeaderSet, Samples, Metadatas, build_trace_header_set, AbstractReader
from .formats import (
    read_ths_from_bin_filenames_list,
    read_ths_from_bin_filenames_pattern,
    read_ths_from_ets_file,
    read_ths_from_trs_file,
    bin_extractor
)
from .formats.bin_format import PaddingMode
import warnings

__all__ = [
    "Trace",
    "TraceHeaderSet",
    "Samples",
    "Metadatas",
    "read_ths_from_bin_filenames_list",
    "read_ths_from_bin_filenames_pattern",
    "read_ths_from_trs_file",
    "bin_extractor",
    "read_ths_from_ets_file",
    "AbstractReader",
    "build_trace_header_set",
    "PaddingMode"
]

# Set default logging handler to avoid "No handler found" warnings.
import logging
from .__version__ import __author__ as AUTHOR, __version__ as VERSION, __author_email__ as AUTHOR_EMAIL  # noqa: F401, N812

logging.getLogger(__name__).addHandler(logging.NullHandler())

# Always display DeprecationWarning by default.
warnings.simplefilter('default', category=DeprecationWarning)



import numpy as _np
from . import abstract_reader


class Samples:
    """Provides a wrapper object around Numpy ndarray for traces samples data manipulation.

    A Samples instance can have 1 or 2 dimensions. It supports a reduced slicing and indexing API.
    ndarray API methods calls are proxied to the wrapped ndarray.

    Attributes:
        array (:class:`ndarray`): wrapped Numpy array.

    """

    SUPPORTED_INDICES_TYPES = (Ellipsis.__class__, slice, int, list, _np.ndarray, range)

    def __init__(self, reader: abstract_reader.AbstractReader, trace_id=None):
        if not isinstance(reader, abstract_reader.AbstractReader):
            raise TypeError(
                "reader must be a subclass of {f}".format(
                    f=abstract_reader.AbstractReader.__name__
                )
            )
        self._trace_id = trace_id
        self._reader = reader
        self._set_ndim()
        # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/35
        self.__array_interface__ = None
        self.__array_struct__ = None

    def __getattr__(self, name):
        arr = self.array
        if hasattr(arr, name):
            return getattr(arr, name)
        else:
            raise AttributeError(f'Attribute {name} does not exists on Samples.')

    def __repr__(self):
        return repr(self.array)

    def __str__(self):
        return repr(self)

    def __len__(self):
        if self.ndim == 1:
            return self._reader.get_trace_size(self._trace_id)
        else:
            return len(self._reader)

    def _set_ndim(self):
        self.ndim = 1 if self._trace_id is not None else 2

    def __getitem__(self, key):
        self._check_indices_type(key)
        traces, frame = self._get_traces_and_frame_from_key(key)
        self._check_traces_dimension(traces_key=traces)
        self._check_frame_dimension(frame=frame)
        if isinstance(traces, int):
            return Samples(reader=self._reader, trace_id=traces)[frame]
        res = self._reader.fetch_samples(traces=traces, frame=frame)
        return self._shaped_result(result=res, traces=traces, frame=frame)

    def _shaped_result(self, result, traces, frame):
        if self.ndim == 1:
            result = result[0]
        if isinstance(frame, int):
            if result.shape[0] == 1:
                return result[0]
            else:
                return result.squeeze()
        else:
            return result

    def _check_frame_dimension(self, frame):
        if isinstance(frame, int):
            frame = [frame]
        if isinstance(frame, range) and len(frame) == 0:
            raise IndexError(f'indexing with length 0 range not allowed.')
        if isinstance(frame, list):
            for f in frame:
                if f >= self._samples_size:
                    raise IndexError(f'index {f} is out of bounds for axis {self._samples_axis} with size {self._samples_size}')

    @property
    def _samples_axis(self):
        return 0 if self.ndim == 1 else 1

    def _check_indices_type(self, indices):
        inds = indices if isinstance(indices, tuple) else (indices, )
        if isinstance(indices, tuple) and len(indices) > self.ndim:
            raise IndexError(f'Samples suport at most {self.ndim} dimensions.')
        for index in inds:
            self._check_index_type(index=index)

    def _check_index_type(self, index):
        if not isinstance(index, self.SUPPORTED_INDICES_TYPES):
            raise IndexError(
                "only {types} are valid indices for Samples".format(
                    types=", ".join([str(t) for t in self.SUPPORTED_INDICES_TYPES]),
                )
            )

    def _get_traces_and_frame_from_key(self, key):
        if self.ndim == 1:
            return [self._trace_id], key
        if not isinstance(key, tuple):
            return self._convert_traces_key_to_iterable(key), ...
        else:
            return self._convert_traces_key_to_iterable(key[0]), key[1]

    def _convert_traces_key_to_iterable(self, traces_key):
        if isinstance(traces_key, (None.__class__, Ellipsis.__class__)):
            return range(len(self._reader))
        elif isinstance(traces_key, (slice, range)):
            stop = traces_key.stop if traces_key.stop is not None and traces_key.stop < len(self._reader) else len(self._reader)
            return range(
                traces_key.start if traces_key.start else 0,
                stop,
                traces_key.step if traces_key.step else 1,
            )
        return traces_key

    def _check_traces_dimension(self, traces_key):
        if isinstance(traces_key, int):
            traces_key = [traces_key]
        if isinstance(traces_key, list):
            for t in traces_key:
                if t > len(self._reader) - 1:
                    raise IndexError(f'index {t} is out of bounds for axis 0 with size {len(self._reader)}')

    @property
    def _samples_size(self):
        if self.ndim == 1:
            return len(self)
        else:
            return len(self[0])

    @property
    def array(self):
        # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/35
        # Having these attributes ensure that numpy.array constructor correctly instantiate array given a Samples instance.
        res = self[:]
        if self.__array_interface__ is None:
            self.__array_struct__ = res.__array_struct__
            self.__array_interface__ = res.__array_interface__
        return res
